# coding: utf-8

# Standard Libraries
import asyncio
import logging

log = logging.getLogger(__name__)


class TaskBase(object):
    worker_threads_num = 1
    active = False

    async def run(self, task):
        self.active = True
        try:
            return await self._run(task)
        finally:
            self.active = False

    async def _run(self, task):
        raise NotImplementedError

    @staticmethod
    async def _run_command(*args):
        """
        Asynchronous run command in subprocess.

        :param *args: command to execute
        :return: tuple (stdout, stderr, exit code)

        Example from: http://asyncio.readthedocs.io/en/latest/subprocess.html
        """
        # Create subprocess
        log.debug("Executing subprocess: %s", " ".join([a for a in args]))
        # pylint: disable=no-value-for-parameter
        process = await asyncio.create_subprocess_exec(
            *args, stderr=asyncio.subprocess.PIPE, stdout=asyncio.subprocess.PIPE)
        # pylint: enable=no-value-for-parameter

        # Status
        log.info('Started: %s (pid = %s)', args, process.pid)

        # Wait for the subprocess to finish
        stdout, stderr = await process.communicate()

        if process.returncode == 0:
            log.debug('Subprocess pid %s succeeded: %s', process.pid, args)
        else:
            log.debug('Subprocess pid %s failed: %s', process.pid, args)

        stdout_str = stdout.decode().strip()
        stderr_str = stderr.decode().strip()

        # Return stdout, stderr, exit code
        return stdout_str, stderr_str, process.returncode



# coding: utf-8

# Standard Libraries
import logging

# Third Party Libraries
from pyannotate_runtime import collect_types

# Dopplerr
from dopplerr.main import main

log = logging.getLogger(__name__)


def main_pyannotate():
    try:
        collect_types.init_types_collection()
        collect_types.resume()
        return main()
    finally:
        log.info("dumping type_info")
        collect_types.pause()
        collect_types.dump_stats('type_info.json')


if __name__ == '__main__':
    main_pyannotate()



# coding: utf-8
"""
Configuration Tree management.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

# Standard Libraries
import argparse
import json
import logging
import os
import sys
from typing import Any
from typing import Dict

# Third Party Libraries
from cfgtree import LOGGER_NAME
from cfgtree.dictxpath import get_node_by_xpath
from cfgtree.dictxpath import set_node_by_xpath
from cfgtree.storages import _ConfigStorageBase

log = logging.getLogger(LOGGER_NAME)
_UNDEFINED = object()


class EnvironmentConfig(object):
    cfgtree: Dict[str, Any] = None
    environ_var_prefix: str = None
    config_storage: _ConfigStorageBase = None

    def __init__(self):
        self._inject_names()

    @staticmethod
    def mkxpath(xpath, name):
        return xpath + "." + name if xpath else name

    def _inject_names(self, root=None, xpath=None):
        """
        Inject configuration item name defined in the cfgtree dict inside each `_Cfg`.
        """
        if root is None:
            if self.cfgtree is None:
                return
            root = self.cfgtree
        # pylint: disable=no-member
        for name, item in root.items():
            if isinstance(item, dict):
                self._inject_names(root=item, xpath=self.mkxpath(xpath, name))
            else:
                item.name = name
                item.xpath = self.mkxpath(xpath, name)
                item.environ_var_prefix = self.environ_var_prefix
                if item.ignore_in_cfg:
                    # log.debug("Create cfg node '%s': ignored (handled later)", item.xpath)
                    continue
                log.debug("Create cfg node: '%s' (name: '%s', cmd line: '%s'), default  : %r",
                          item.xpath, item.name, item.long_param, item.safe_value)
        # pylint: enable=no-member

    def set_cfg_value(self, xpath, value):
        """
        Set a value in cfgtree.
        """
        set_node_by_xpath(self.cfgtree, xpath, value, extend=True, setter_attr="set_value")

    def get_cfg_value(self, xpath, default=None):
        """
        Get a value from cfgtree.
        """
        return get_node_by_xpath(self.cfgtree, xpath, default=default).value

    def find_configuration_values(self):
        self._load_configuration()
        self._load_environment_variables("", self.cfgtree)
        self._load_cmd_line_arg()

    def _load_configuration(self):
        log.debug("Looking for configuration")
        self.config_storage.find_config_storage()
        bare_cfg = self.config_storage.get_bare_config_dict()
        self._load_cfg_dict(bare_cfg)

    def save_configuration(self):
        log.debug("Saving configuration file")
        bare_cfg = self._dict(safe=False)
        self.config_storage.save_bare_config_dict(bare_cfg)

    def _load_cfg_dict(self, cfg, xpath=None):
        for k, v in cfg.items():
            xp = self.mkxpath(xpath, k)
            if isinstance(v, dict):
                self._load_cfg_dict(v, xp)
            else:
                try:
                    self.set_cfg_value(xp, v)
                except KeyError:
                    log.error("Unable to load value '%s' from configuration file, "
                              "no matching item in configuration tree (invalid '%s')", k, xp)

    def _load_environment_variables(self, xpath, root):
        """
        Inject value from environment variable.
        """
        for name, item in root.items():
            if isinstance(item, dict):
                self._load_environment_variables(self.mkxpath(xpath, name), item)
            else:
                if item.environ_var_name in os.environ:
                    if item.ignore_in_cfg:
                        log.debug("Ignoring environment variable %s", item.environ_var_name)
                    val = item.read_environ_var()
                    log.debug("Found environment variable '%s': %s (conf: %s)",
                              item.environ_var_name, val, item.xpath)
                    item.value = val

    def _load_cmd_line_arg(self):
        """
        Inject parameters provider by the user in the command line.
        """
        argv = sys.argv[1:]

        parser = argparse.ArgumentParser()
        self._inject_cfg_in_parser(parser)
        args = parser.parse_args(args=argv)
        for k, v in vars(args).items():
            cfg = self._find_cfg_for_cmd_line_name(k)
            if v is _UNDEFINED:
                continue
            cfg.value = v
            if cfg.ignore_in_cfg or cfg.ignore_in_args:
                log.debug("Ignoring command line parameter %s", cfg.long_param)
            log.debug("Found command line parameter '%s': %s (conf: %s)", cfg.long_param,
                      cfg.safe_value, cfg.xpath)

    def _find_cfg_for_cmd_line_name(self, cmd_line_name, root=None):
        if root is None:
            root = self.cfgtree
        for v in root.values():
            if isinstance(v, dict):
                f = self._find_cfg_for_cmd_line_name(cmd_line_name, root=v)
                if f:
                    return f
            else:
                if v.cmd_line_name == cmd_line_name:
                    return v

    def _inject_cfg_in_parser(self, parser, xpath=None, root=None):
        """
        Configure the argument parser according to cfgtree.
        """
        if root is None:
            root = self.cfgtree
        # pylint: disable=no-member
        for name, item in root.items():
            if isinstance(item, dict):
                self._inject_cfg_in_parser(parser, xpath=self.mkxpath(xpath, name), root=item)
            else:
                if item.ignore_in_args:
                    continue
                args = item.get_cmd_line_params()
                kwargs = {
                    "action": item.action,
                    "dest": item.cmd_line_name,
                    "help": item.help_str,
                    "default": _UNDEFINED,
                }
                nargs = item.n_args
                dbg_infos = ["arg '{}'".format(item.long_param), "dest '{}'".format(kwargs['dest'])]
                if nargs:
                    kwargs["nargs"] = nargs
                    dbg_infos.append("nargs '{}'".format(str(nargs)))
                metavar = item.metavar
                if metavar:
                    kwargs["metavar"] = metavar
                    dbg_infos.append("metavar '{}'".format(metavar))
                if item.arg_type is not None:
                    kwargs["type"] = item.arg_type
                    dbg_infos.append("type '{}'".format(item.arg_type))
                log.debug("parser %s", ", ".join(dbg_infos))
                parser.add_argument(*args, **kwargs)
        # pylint: enable=no-member

    def _dict(self, root=None, safe=False):
        """
        Return the configuration as a dictionnary.
        """
        if root is None:
            root = self.cfgtree
        d = {}
        # pylint: disable=no-member
        for name, item in root.items():
            if isinstance(item, dict):
                d[name] = self._dict(root=item, safe=safe)
            else:
                if item.ignore_in_cfg:
                    continue
                elif safe:
                    d[name] = item.safe_value
                else:
                    d[name] = item.value
        # pylint: enable=no-member
        return d

    def json(self, safe=False):
        return json.dumps(self._dict(safe=safe), sort_keys=True, indent=4, separators=(',', ': '))



# coding: utf-8

# Standard Libraries
import glob
import logging
import os
from pathlib import Path

# Dopplerr
from dopplerr.config import DopplerrConfig
from dopplerr.db import DopplerrDb
from dopplerr.descriptors.media_types import VIDEO_FILES_EXT
from dopplerr.singleton import singleton
from dopplerr.status import DopplerrStatus
from dopplerr.tasks.queued import QueuedTask
from dopplerr.tasks.subtasks.subliminal import SubliminalSubDownloader

log = logging.getLogger(__name__)


@singleton
class DownloadSubtitleTask(QueuedTask):
    name = "Download Subtitle Task"

    async def _run(self, task):
        log.debug("Starting Download subtitle for task")
        res = task
        if not res.candidates:
            DopplerrDb().insert_event("error", "event handled but no candidate found")
            log.debug("event handled but no candidate found")
            res.failed("event handled but no candidate found")
            return res

        for candidate in res.candidates:
            await self._process_candidate(candidate, res)
        return res

    @staticmethod
    def search_file(root_dir, base_name):
        # This won't work with python < 3.5
        found = []
        base_name = glob.escape(base_name)
        beforext, _, ext = base_name.rpartition('.')
        if ext.lower() in VIDEO_FILES_EXT:
            protected_path = os.path.join(root_dir, "**", "*" + beforext + "*" + ext)
        else:
            protected_path = os.path.join(root_dir, "**", "*" + beforext + "*")
        protected_path = protected_path
        log.debug("Searching %r", protected_path)
        for filename in glob.iglob(protected_path, recursive=True):
            log.debug("Found: %s", filename)
            found.append(filename)
        return found

    async def _process_candidate(self, candidate, res):
        log.info(
            "Searching episode '%s' from series '%s'. Filename: %s",
            candidate.get("episode_title"),
            candidate.get("series_title"),
            candidate.get("scenename"),
        )

        candidate_files = self._search_candidate_files(candidate, res)
        if not candidate_files:
            return

        self._refresh_db_media(candidate, candidate_files[0])

        videos = self.filter_video_files(candidate_files, res)
        if not videos:
            return

        subtitles_info = await self.download_sub(videos, res)
        res.subtitles = subtitles_info
        if subtitles_info:
            res.successful("download successful")
            DopplerrDb().insert_event("subtitles",
                                      "subtitles fetched: {}".format(", ".join([
                                          "{} (lang: {}, source: {})".format(
                                              s.get("filename"),
                                              s.get("language"),
                                              s.get("provider"),
                                          ) for s in subtitles_info
                                      ])))
        else:
            DopplerrDb().insert_event("subtitles", "no subtitle found for: {}".format(
                ", ".join([Path(f).name for f in candidate_files])))
            res.failed("no subtitle found")

    @staticmethod
    def _refresh_db_media(candidate, media_filename):
        DopplerrDb().update_series_media(
            series_title=candidate.get("series_title"),
            tv_db_id=candidate.get("tv_db_id"),
            season_number=candidate.get("season_number"),
            episode_number=candidate.get("episode_number"),
            episode_title=candidate.get("episode_title"),
            quality=candidate.get("quality"),
            video_languages=None,
            media_filename=media_filename,
            dirty=True)

    def _search_candidate_files(self, candidate, res):
        candidate_files = self.search_file(candidate['root_dir'], candidate['scenename'])
        log.debug("All found files: %r", candidate_files)
        if not candidate_files:
            res.failed("candidates defined in request but no video file found on disk")
            DopplerrDb().insert_event("subtitles", "No video file found on disk "
                                      "after sonarr notification")
            return []
        return candidate_files

    @staticmethod
    def filter_video_files(candidate_files, res):
        log.info("Searching and downloading missing subtitles for: %r", candidate_files)
        res.processing("downloading missing subtitles")
        videos = SubliminalSubDownloader.filter_video_files(candidate_files)
        log.info("Video files: %r", videos)
        if not videos:
            log.debug("No subtitle to download")
            res.failed("no video file found")
            return
        return videos

    async def download_sub(self, videos, res):
        res.processing("fetching best subtitles")
        log.info("fetching subtitles...")
        subtitles = []
        try:
            subliminal = SubliminalSubDownloader()
            provider_configs = DopplerrStatus().subliminal_provider_configs
            languages = DopplerrConfig().get_cfg_value("subliminal.languages")
            subtitles = await subliminal.download_sub(
                videos, languages, provider_configs=provider_configs)
        except Exception as e:
            log.exception("subliminal raised an exception")
            res.failed("subliminal exception")
            res.exception = repr(e)
            return res

        subtitles_info = []
        for vid in videos:
            log.info("Found subtitles for %s:", vid)
            for sub in subtitles[vid]:
                log.info("  %s from %s", sub.language, sub.provider_name)
                subtitles_info.append({
                    "language": str(sub.language),
                    "provider": sub.provider_name,
                    "filename": subliminal.get_subtitle_path(vid.name, language=sub.language)
                })
            subliminal.save_subtitles(vid, subtitles[vid])

        return subtitles_info



# coding: utf-8

# Standard Libraries
import logging

# Dopplerr
from dopplerr.notifications_types._base import _SeriesNotificationBase

log = logging.getLogger(__name__)


# pylint: disable=duplicate-code
class SubtitleFetchedNotification(_SeriesNotificationBase):
    notification_type = "fetched"
    notification_title = "Episode Subtitles Fetched"

    @property
    def one_liner(self):
        return ("{e[series_title]} - {e[season_number]}x{e[episode_number]} - "
                "{e[episode_title]} [{e[quality]}] - Lang: {e[video_languages]} - "
                "Subtitles: {subtitles_languages}".format(
                    e=self.series_episode_info,
                    subtitles_languages=",".join(self.series_episode_info['subtitles_languages']),
                ))



# coding: utf-8

# Standard Libraries
from enum import Enum
from json import JSONEncoder
from json import dumps


def _pretty_kw():
    return {
        "sort_keys": True,
        "indent": 4,
        "separators": (',', ': '),
    }


class _EnumEncoder(JSONEncoder):

    def default(self, obj):  # pylint: disable=arguments-differ,method-hidden
        if isinstance(obj, Enum):
            return obj.name
        return JSONEncoder.default(self, obj)


def safe_dumps(data):
    return dumps(data, cls=_EnumEncoder, **_pretty_kw())



# coding: utf-8

# Standard Libraries
import logging
from enum import Enum

# Dopplerr
from dopplerr import json

log = logging.getLogger(__name__)


class RequestStatus(Enum):
    UNHANDLED = "unhandled"
    PROCESSING = "processing"
    SUCCESSFUL = "successful"
    FAILED = "failed"


class Response(object):

    def __init__(self):
        self.res = {}
        self.__update_status(RequestStatus.UNHANDLED)

    def processing(self, message=None):
        self.__update_status(RequestStatus.PROCESSING, message=message)

    def failed(self, message):
        log.error(message)
        self.__update_status(RequestStatus.FAILED, message=message.lower())

    def unhandled(self, message):
        log.info("Filtered out event: %s", message)
        self.res.setdefault("result", {})["status"] = RequestStatus.UNHANDLED
        self.res.setdefault("result", {})["message"] = message.lower()

    @property
    def is_unhandled(self):
        return self.res.get("result", {}).get("status") == RequestStatus.UNHANDLED

    @property
    def is_failed(self):
        return self.res.get("result", {}).get("status") == RequestStatus.FAILED

    def __update_status(self, status, message=None):
        self.res.setdefault("result", {})['status'] = status
        if message is not None:
            self.res['result']["message"] = message
        elif "message" in self.res:
            del self.res['result']["message"]

    def successful(self, message=None):
        return self.__update_status(RequestStatus.SUCCESSFUL, message=message)

    @property
    def is_successful(self):
        return self.res.get("result", {}).get("status") == RequestStatus.SUCCESSFUL

    def to_dict(self):
        """
        Return json-able dictionary.
        """
        return self._to_dict(self.res)

    def _to_dict(self, dat):
        r = {}
        for k, v in dat.items():
            if isinstance(v, Enum):
                v = v.name
            elif isinstance(v, dict):
                v = self._to_dict(v)
            r[k] = v
        return r

    def to_json(self):
        return json.safe_dumps(self.res)

    @property
    def request_type(self):
        return self.res.get("request", {}).get("type", None)

    @request_type.setter
    def request_type(self, thetype):
        self.res.setdefault("request", {})["type"] = thetype

    @property
    def request_event(self):
        return self.res.get("request", {}).get("event", None)

    @request_event.setter
    def request_event(self, event):
        self.res.setdefault("request", {})["event"] = event

    @property
    def exception(self):
        return self.res.get("result", {}).get("exception", None)

    @exception.setter
    def exception(self, exception):
        self.res.setdefault("result", {})["exception"] = exception


class UnhandledResponse(Response):

    def __init__(self, request_type, message, *args, **kwargs):
        super(UnhandledResponse, self).__init__(*args, **kwargs)
        self.request_type = request_type
        self.unhandled(message)



# coding: utf-8

# Standard Libraries
import asyncio
import logging

# Dopplerr
from dopplerr.tasks.base import TaskBase

log = logging.getLogger(__name__)


class QueuedTask(TaskBase):

    """
    Task that should be queued.

    Ie, where two similar cannot be executed at the same time,
    but to an underlying shared resource.

    Usage:

    - inherit from `QueuedTask` and implement `name` member and `_run` methods

        class MyTask(QueuedTask):
            name = "MyTask"

            async def _run(self, task):
                ...

    """

    _queue: asyncio.Queue = None
    _result: asyncio.Queue = None
    _started: bool = False
    _consumer: asyncio.Future = None
    name: str = "Unnamed queue"
    FIRE_AND_FORGET = 0
    WAIT_FOR_RESULT = 1

    def __init__(self):
        self._queue = asyncio.Queue()
        self._result = asyncio.Queue()

    async def _loop_queue(self):
        assert self._queue
        while True:
            # log.debug("QueuedTask %s: waits for something in the task queue", self.name)
            fire_and_forget, inputs = await self._queue.get()

            try:
                self.active = True
                # log.debug("QueuedTask %s: starts execution: %r", self.name, inputs)
                result = await self._run(inputs)
                # log.debug("QueuedTask %s: ends execution: %r (result: %r)", self.name, inputs,
                #           result)
                if fire_and_forget == QueuedTask.WAIT_FOR_RESULT:
                    # log.debug("QueuedTask %s: task %s is 'wait for result', posting result %r",
                    #           self.name, inputs, result)
                    await self._result.put(result)
            finally:
                self.active = False
                self._queue.task_done()

    async def _run(self, task):
        raise NotImplementedError

    async def _wait_next_result(self):
        assert self.started
        # log.debug("QueuedTask %s: waiting for result", self.name)
        r = await self._result.get()
        # log.debug("QueuedTask %s: previous task returned a result: %s", self.name, r)
        self._result.task_done()
        return r

    def start(self):
        assert self._queue
        if self.stopped:
            # log.debug("QueuedTask %s: starting queue", self.name)
            self._consumer = asyncio.ensure_future(self._loop_queue())

    def stop(self):
        assert self._queue
        # log.debug("QueuedTask %s: stopping queue", self.name)
        if self._consumer:
            self._consumer.cancel()
        self._consumer = None

    async def fire_and_forget(self, task):
        assert self._queue
        # log.debug("QueuedTask %s: posting new execution request: %r", self.name, task)
        await self._queue.put((QueuedTask.FIRE_AND_FORGET, task))

    async def run(self, task):
        return await self.run_and_wait(task)

    async def run_and_wait(self, task):
        assert self._queue
        # log.debug("QueuedTask %s: posting 'Run and resume' job: %r", self.name, task)
        await self._queue.put((QueuedTask.WAIT_FOR_RESULT, task))
        # log.debug("QueuedTask %s: task %s posted, waiting for result", self.name, task)
        return await self._wait_next_result()

    async def join(self):
        assert self._queue
        await self._queue.join()

    @property
    def stopped(self):
        return not self.started

    @property
    def started(self):
        return self._consumer



# coding: utf-8

# Standard Libraries
import logging
import os
from datetime import timedelta
from typing import List

# Third Party Libraries
from babelfish import Language
from subliminal import Episode
from subliminal import Movie
from subliminal import Video
from subliminal import download_best_subtitles
from subliminal import refine
from subliminal import refiner_manager
from subliminal import region
from subliminal import save_subtitles
from subliminal.cli import MutexLock
from subliminal.subtitle import get_subtitle_path

# Dopplerr
from dopplerr.descriptors.series import SeriesEpisodeInfo
from dopplerr.descriptors.series import SeriesEpisodeUid
from dopplerr.tasks.threaded import ThreadedTask

log = logging.getLogger(__name__)


class SubliminalSubDownloader(ThreadedTask):
    worker_threads_num = 1

    async def _run(self, res):
        raise NotImplementedError

    @staticmethod
    def initialize_db():
        log.info("Initializing Subliminal cache...")
        region.configure(
            'dogpile.cache.dbm',
            expiration_time=timedelta(days=30),
            arguments={
                'filename': 'cachefile.dbm',
                'lock_factory': MutexLock
            })

    # async def download_sub_by_subproc(self, videos, languages, provider_configs):
    #     subl_cmd = ["subliminal"]
    #     print(provider_configs)
    #     for provider_name, provider_config in provider_configs.items():
    #         subl_cmd.extend([
    #             "--{}".format(provider_name),
    #             provider_config['username'],
    #             provider_config['password'],
    #         ])
    #     subl_cmd.extend(["download"])
    #     for l in languages:
    #         subl_cmd.extend(["--language", l])
    #     subl_cmd.extend(videos)
    #     subl_cmd.extend(["-vvv"])
    #     stdout, stderr, code = await self._run_command(*subl_cmd)
    #     log.debug(stdout)
    #     log.error(stderr)
    #     log.error(code)
    #     if "Downloaded 0 subtitle" in stdout:
    #         log.error("No subtitle downloaded")
    #     raise NotImplementedError

    async def download_sub(self, videos, languages, provider_configs):
        return await self._run_in_thread(
            download_best_subtitles,
            videos, {Language(l)
                     for l in languages},
            provider_configs=provider_configs)

    @staticmethod
    def filter_video_files(files):
        videos = []
        for fil in files:
            _, ext = os.path.splitext(fil)
            if ext in [".jpeg", ".jpg", ".nfo", ".srt", ".sub", ".nbz"]:
                log.debug("Ignoring %s because of extension: %s", fil, ext)
                continue
            videos.append(Video.fromname(fil))
        return videos

    @staticmethod
    def get_subtitle_path(video_file, language):
        return get_subtitle_path(video_file, language=language)

    @staticmethod
    def save_subtitles(video, subtitle_info):
        return save_subtitles(video, subtitle_info)


class RefineVideoFileTask(ThreadedTask):

    async def refine_file(self, video_file):
        return await self._run_in_thread(self._refine_file, video_file)

    @staticmethod
    def _refine_file(video_file) -> List[SeriesEpisodeInfo]:
        log.debug("Refining file %s", video_file)
        try:
            video = Video.fromname(video_file)
        except ValueError:
            log.error("Cannot guess video file type from: %s", video_file)
            return []
        refiner = sorted(refiner_manager.names())
        refine(video, episode_refiners=refiner, movie_refiners=refiner)
        log.debug("refine result: %r", video)
        if isinstance(video, Episode):
            log.debug("series: %s", video.series)
            log.debug("season: %s", video.season)
            if not video.season:
                log.error("No season defined !")
                return []
            if isinstance(video.season, list):
                video.season = video.season[0]
                log.error("Multi season found, only using first one: %s", video.season)
            log.debug("episode: %s", video.episode)
            log.debug("title: %s", video.title)
            log.debug("series_tvdb_id: %s", video.series_tvdb_id)
            r = []
            # Support for double episode
            if not isinstance(video.episode, list):
                video.episode = [video.episode]
            for video_episode in video.episode:
                r.append(
                    SeriesEpisodeInfo(
                        series_episode_uid=SeriesEpisodeUid(
                            tv_db_id=video.series_tvdb_id,
                            season_number=video.season,
                            episode_number=video_episode,
                        ),
                        series_title=video.series,
                        episode_title=video.title,
                        quality=None,
                        video_languages=None,
                        subtitles_languages=None,
                        media_filename=video_file,
                        dirty=True,
                    ))
            return r
        elif isinstance(video, Movie):
            log.debug("movie: %s", video.title)
        return []



# coding: utf-8

# Standard Libraries
import logging
from pathlib import Path

# Third Party Libraries
from sanic import Sanic
from sanic_transmute import add_swagger

# Dopplerr
from dopplerr.api.v1 import add_api_blueprints
from dopplerr.config import DopplerrConfig
from dopplerr.status import DopplerrStatus
from dopplerr.tasks.manager import DopplerrTasksManager

log = logging.getLogger(__name__)


async def init_in_sanic_loop():
    DopplerrTasksManager().start()


async def deinit_in_sanic_loop():
    DopplerrTasksManager().stop()


def listen():
    app = Sanic(__name__, log_config=None)
    add_api_blueprints(app)
    add_swagger(app, "/api/v1/swagger.json", "/api/v1/")

    DopplerrStatus().healthy = True
    for fi in Path(DopplerrConfig().get_cfg_value("general.frontenddir")).iterdir():
        app.static('/' + fi.name if fi.name != "index.html" else '/', fi.resolve().as_posix())

    @app.listener('before_server_start')
    async def before_start(_app, _loop):  # pylint: disable=unused-variable
        await init_in_sanic_loop()

    @app.listener('after_server_stop')
    async def after_stop(_app, _loop):  # pylint: disable=unused-variable
        await deinit_in_sanic_loop()

    app.run(host='0.0.0.0', port=int(DopplerrConfig().get_cfg_value("general.port")))



# coding: utf-8

# Standard Libraries
import io
import logging

# Third Party Libraries
import aiofiles
from babelfish import Language

# Dopplerr
from dopplerr import DOPPLERR_VERSION
from dopplerr.config import DopplerrConfig
from dopplerr.singleton import singleton

log = logging.getLogger(__name__)


@singleton
class DopplerrStatus(object):

    """
    Contain current status of the application and derived values from `DopplerrConfig`.
    """

    def __init__(self):
        self.healthy = False
        self.sqlite_db_path = None
        self.subliminal_provider_configs = None
        self.previous_version = None

    def refresh_from_cfg(self):
        """
        Refresh derived values from cfg.
        """
        cfg = DopplerrConfig()
        if not cfg.get_cfg_value("general.port"):
            log.fatal("No port defined !")
            raise Exception("No port defined")
        if not cfg.get_cfg_value("general.frontenddir"):
            log.fatal("No frontend dir defined")
            raise Exception("No frontend dir defined")
        self.subliminal_provider_configs = self._build_subliminal_provider_cfgs()

        languages = cfg.get_cfg_value("subliminal.languages")
        if not languages:
            raise Exception("No languages defined")
        if any(not x for x in languages):
            raise Exception("Bad languages: {!r}".format(languages))

        if not self._check_languages(languages):
            raise Exception("Bad language defined")

        if self.previous_version is None:
            self.previous_version = cfg.get_cfg_value("general.version")
            cfg.set_cfg_value("general.version", DOPPLERR_VERSION)

    @property
    def has_minor_version_changed(self):
        if not self.previous_version:
            return True
        major1, _, minor_patch1 = self.previous_version.partition('.')
        major2, _, minor_patch2 = DOPPLERR_VERSION.partition('.')
        minor1, _, _patch1 = minor_patch1.partition('.')
        minor2, _, _patch2 = minor_patch2.partition('.')
        return major1 != major2 or minor1 != minor2

    def _build_subliminal_provider_cfgs(self):
        cfg = DopplerrConfig()
        provider_configs = {}
        provider_names = [
            "addic7ed",
            "legendastv",
            "opensubtitles",
            "subscenter",
        ]
        for provider_name in provider_names:
            if cfg.get_cfg_value("subliminal.{}.enabled".format(provider_name)):
                provider_configs[provider_name] = {
                    'username': cfg.get_cfg_value("subliminal.{}.user".format(provider_name)),
                    'password': cfg.get_cfg_value("subliminal.{}.password".format(provider_name)),
                }
                log.debug("Using %s username: %s", provider_name,
                          provider_configs[provider_name]['username'])
        return provider_configs

    @staticmethod
    def _check_languages(languages):
        failed = False
        for l in languages:
            try:
                Language(l)
            except ValueError:
                failed = True
                logging.critical("Invalid language: %r", l)
        if failed:
            return False
        return True

    async def get_logs(self, limit=100):
        """
        Get `limit` lines of logs in reverse order from the end of the file.
        """
        logfile = DopplerrConfig().get_cfg_value("general.logfile")
        if not logfile:
            return
        logs = []
        i = 0
        async with aiofiles.open(logfile) as fp:
            async for line in self._reverse_read_lines(fp):
                try:
                    i += 1
                    if i > limit:
                        break
                    if not line:
                        continue
                    splited_line = line.split("::")
                    if len(splited_line) < 4:
                        continue
                    dat = splited_line[0].strip()
                    level = splited_line[1].strip()
                    logger = splited_line[2].strip()
                    message = splited_line[3].strip()
                    logs.append({
                        'timestamp': dat,
                        'level': level,
                        'logger': logger,
                        'message': message,
                    })
                finally:
                    pass
        return logs

    @staticmethod
    async def _reverse_read_lines(fp, buf_size=8192):  # pylint: disable=invalid-name
        """
        Async generator that returns the lines of a file in reverse order.

        ref: https://stackoverflow.com/a/23646049/8776239
        and: https://stackoverflow.com/questions/2301789/read-a-file-in-reverse-order-using-python
        """
        segment = None  # holds possible incomplete segment at the beginning of the buffer
        offset = 0
        await fp.seek(0, io.SEEK_END)
        file_size = remaining_size = await fp.tell()
        while remaining_size > 0:
            offset = min(file_size, offset + buf_size)
            await fp.seek(file_size - offset)
            buffer = await fp.read(min(remaining_size, buf_size))
            remaining_size -= buf_size
            lines = buffer.splitlines(True)
            # the first line of the buffer is probably not a complete line so
            # we'll save it and append it to the last line of the next buffer
            # we read
            if segment is not None:
                # if the previous chunk starts right from the beginning of line
                # do not concat the segment to the last line of new chunk
                # instead, yield the segment first
                if buffer[-1] == '\n':
                    # print 'buffer ends with newline'
                    yield segment
                else:
                    lines[-1] += segment
                    # print 'enlarged last line to >{}<, len {}'.format(lines[-1], len(lines))
            segment = lines[0]
            for index in range(len(lines) - 1, 0, -1):
                l = lines[index]
                if l:
                    yield l
        # Don't yield None if the file was empty
        if segment is not None:
            yield segment



# coding: utf-8

# Standard Libraries
import asyncio
import logging
import os

# Dopplerr
from dopplerr.config import DopplerrConfig
from dopplerr.db import DopplerrDb
from dopplerr.descriptors.media_types import VIDEO_FILES_EXT
from dopplerr.descriptors.series import SeriesEpisodeInfo
from dopplerr.singleton import singleton
from dopplerr.tasks.periodic import PeriodicTask
from dopplerr.tasks.subtasks.subliminal import RefineVideoFileTask

log = logging.getLogger(__name__)

SPEED_LIMIT = 10
SPEED_WAIT_SEC = 0.1


@singleton
class DiskScanner(PeriodicTask):
    job_id = 'scan_disk'
    seconds = 0
    minutes = 0
    hours = None
    enable_cfg = 'scanner.enable'

    def init(self):
        self.hours = DopplerrConfig().get_cfg_value('scanner.interval_hours')

    async def _run(self):
        basedir = DopplerrConfig().get_cfg_value('general.basedir')
        mapping = DopplerrConfig().get_cfg_value('general.mapping')
        media_dirs = []
        for mapp in mapping:
            _src, _, dst = mapp.partition('=')
            media_dirs.append(dst)
        log.debug("Scanning %s with media directories: %r", basedir, media_dirs)
        await self._scan(basedir, media_dirs=media_dirs)

    async def _scan(self, root, media_dirs=None):
        i = 0
        with os.scandir(root) as it:
            for entry in it:
                i += 1
                if i > SPEED_LIMIT:
                    # this allows the event loop to update
                    await asyncio.sleep(SPEED_WAIT_SEC)
                    i = 0
                if self.stopped or self.interrupted:
                    return
                if media_dirs:
                    if entry.name in media_dirs:
                        await self._scan(entry.path, media_dirs=None)
                elif not entry.name.startswith('.'):
                    if entry.is_dir(follow_symlinks=False):
                        await self._scan(entry.path, media_dirs=None)
                    elif entry.name.rpartition('.')[2] in VIDEO_FILES_EXT:
                        await self._refresh_video(entry.path)

    async def _refresh_video(self, filepath):
        if DopplerrDb().media_exists(filepath):
            log.info("Already existing video file found: %s", filepath)
            return
        log.info("Unknown Video file found: %s", filepath)
        refined_lst = await RefineVideoFileTask().refine_file(filepath)
        if not refined_lst:
            log.error("Cannot refine file: %s", filepath)
            return
        if not isinstance(refined_lst, list):
            refined_lst = [refined_lst]
        for refined in refined_lst:
            if isinstance(refined, SeriesEpisodeInfo):
                DopplerrDb().update_series_media(
                    series_title=refined.series_title,
                    tv_db_id=refined.series_episode_uid.tv_db_id,
                    season_number=refined.series_episode_uid.season_number,
                    episode_number=refined.series_episode_uid.episode_number,
                    episode_title=refined.episode_title,
                    quality=refined.quality,
                    video_languages=refined.video_languages,
                    media_filename=refined.media_filename,
                    dirty=refined.dirty,
                )
                DopplerrDb().insert_event("availability", "Available: {} - {}x{} - {}.".format(
                    refined.series_title,
                    refined.series_episode_uid.season_number,
                    refined.series_episode_uid.episode_number,
                    refined.episode_title,
                ))
            else:
                log.error("Unsupported refined video type: %r", refined)
                return

    @property
    def interval_hours(self):
        return self.hours



# coding: utf-8

# Standard Libraries
import logging

# Third Party Libraries
from sanic import Blueprint
from sanic_transmute import add_route
from sanic_transmute import describe
from schematics.models import Model
from schematics.types import DateType
from schematics.types import ListType
from schematics.types import ModelType
from schematics.types import StringType

# Dopplerr
from dopplerr.db import DopplerrDb

log = logging.getLogger(__name__)

bp = Blueprint('series', url_prefix="/api/v1/medias/series")


class Media(Model):
    timestamp = DateType()
    type = StringType()
    message = StringType()


class Series(Model):
    medias = ListType(ModelType(Media))


@describe(paths="/all", methods="GET")
async def medias_series():
    res = {"medias": DopplerrDb().get_medias_series()}
    return res


add_route(bp, medias_series)



# coding: utf-8

# Standard Libraries
import logging

# Third Party Libraries
from sanic import Blueprint
from sanic_transmute import add_route
from sanic_transmute import describe
from schematics.models import Model

# Dopplerr
from dopplerr.downloader import DopplerrDownloader

log = logging.getLogger(__name__)


class RequestAnswer(Model):
    pass


bp = Blueprint('medias', url_prefix="/api/v1")


@describe(paths="/medias/", methods="GET")
async def fullscan(request) -> RequestAnswer:
    content = request.json
    logging.debug("Fullscan request: %r", content)
    res = DopplerrDownloader().process_fullscan(content)
    res = "Unimplemented"
    return res


add_route(bp, fullscan)



# coding: utf-8

import platform

import pbr.version

__version__ = pbr.version.VersionInfo('dopplerr').release_string()
VERSION = __version__
DOPPLERR_VERSION = __version__

APSSCHEDULER_VERSION = pbr.version.VersionInfo('apscheduler').release_string()
PYTHON_VERSION = platform.python_version()
SANIC_VERSION = pbr.version.VersionInfo('sanic').release_string()

__all__ = [
    '__version__',
    'APSSCHEDULER_VERSION',
    'DOPPLERR_VERSION',
    'PYTHON_VERSION',
    'SANIC_VERSION',
    'VERSION',
]



# coding: utf-8

# Standard Libraries
import logging

# Third Party Libraries
from sanic import Blueprint
from sanic_transmute import add_route
from sanic_transmute import describe
from schematics.models import Model
from schematics.types import BooleanType
from schematics.types import IntType
from schematics.types import ListType
from schematics.types import ModelType
from schematics.types import StringType

# Dopplerr
from dopplerr import APSSCHEDULER_VERSION
from dopplerr import DOPPLERR_VERSION
from dopplerr import PYTHON_VERSION
from dopplerr import SANIC_VERSION
from dopplerr.config import DopplerrConfig
from dopplerr.status import DopplerrStatus
from dopplerr.tasks.disk_scanner import DiskScanner
from dopplerr.tasks.manager import DopplerrTasksManager

log = logging.getLogger(__name__)


class SubtitleDownloader(Model):
    active = IntType()
    started = IntType()


class DiscScanner(Model):
    interval_hours = IntType()
    next_run_time = StringType()
    active = IntType()
    started = IntType()


class TaskStatus(Model):
    downloader = StringType()
    background_tasks = IntType()
    subtitle_downloader = ModelType(SubtitleDownloader)
    disc_scanner = ModelType(DiscScanner)


class Health(Model):
    healthy = BooleanType()
    languages = ListType(StringType())
    mapping = ListType(StringType())
    version = StringType()


class Version(StringType):
    pass


class Versions(Model):
    dopplerr = StringType()
    apscheduler = StringType()
    sanic = StringType()
    python = StringType()


class Log(Model):
    timestamp = StringType()
    level = StringType()
    message = StringType()
    logger = StringType()


Logs = ListType(ModelType(Log))

bp = Blueprint('status', url_prefix="/api/v1")


@describe(paths="/health")
async def health() -> Health:
    res_health = {
        "healthy": DopplerrStatus().healthy,
        "languages": DopplerrConfig().get_cfg_value("subliminal.languages"),
        "mapping": DopplerrConfig().get_cfg_value("general.mapping"),
        "version": DOPPLERR_VERSION,
    }
    return res_health


add_route(bp, health)


@describe(paths="/tasks/status")
async def tasks_status() -> TaskStatus:
    res_health = DopplerrTasksManager().status()
    return res_health


add_route(bp, tasks_status)


@describe(paths="/tasks/scanner/start", methods=["POST"])
async def start_scanner() -> TaskStatus:
    await DiskScanner().force_start()
    return "OK"


add_route(bp, start_scanner)


@describe(paths="/version")
async def api_version() -> Version:
    return DOPPLERR_VERSION


add_route(bp, api_version)


@describe(paths="/versions")
async def api_versions() -> Versions:
    return {
        "dopplerr": DOPPLERR_VERSION,
        "apscheduler": APSSCHEDULER_VERSION,
        "sanic": SANIC_VERSION,
        "python": PYTHON_VERSION,
    }


add_route(bp, api_versions)


@describe(paths="/logs", query_parameters=['limit'])
async def api_log_100(limit: int = 100) -> Logs:
    return await DopplerrStatus().get_logs(limit)


add_route(bp, api_log_100)



# coding: utf-8

import dopplerr.api.v1.config
import dopplerr.api.v1.events
import dopplerr.api.v1.medias
import dopplerr.api.v1.notify
import dopplerr.api.v1.series
import dopplerr.api.v1.status


def add_api_blueprints(app):
    app.blueprint(dopplerr.api.v1.config.bp)
    app.blueprint(dopplerr.api.v1.events.bp)
    app.blueprint(dopplerr.api.v1.medias.bp)
    app.blueprint(dopplerr.api.v1.notify.bp)
    app.blueprint(dopplerr.api.v1.series.bp)
    app.blueprint(dopplerr.api.v1.status.bp)



# coding: utf-8

# Standard Libraries
import json
import logging
import os
import sys
from pathlib import PosixPath

# Third Party Libraries
from cfgtree import LOGGER_NAME

log = logging.getLogger(LOGGER_NAME)


class _ConfigStorageBase(object):

    def find_config_storage(self):
        raise NotImplementedError

    def get_bare_config_dict(self):
        raise NotImplementedError

    def save_bare_config_dict(self, bare_cfg):
        raise NotImplementedError


class JsonFileConfigStorage(_ConfigStorageBase):
    json_configstorage_default_filename: str = None
    json_configstorage_environ_var_name: str = None
    json_configstorage_short_param_name: str = None
    json_configstorage_long_param_name: str = None

    __resolved_config_file = None
    __bare_config_dict = None

    def find_config_storage(self):
        configfile = self.json_configstorage_default_filename
        if self.json_configstorage_environ_var_name in os.environ:
            configfile = os.environ[self.json_configstorage_environ_var_name]
            log.debug("%s defined: %s", self.json_configstorage_environ_var_name, configfile)
        for i in range(len(sys.argv)):
            good = []
            if self.json_configstorage_short_param_name:
                good.append(self.json_configstorage_short_param_name)
            if self.json_configstorage_long_param_name:
                good.append(self.json_configstorage_long_param_name)
            if sys.argv[i] in good:
                if i == len(sys.argv):
                    raise Exception("No value given to {}".format(" or ".join(good)))
                configfile = sys.argv[i + 1]
                log.debug("%s defined: %s", " or ".join(good), configfile)
                break
        config_file_path = PosixPath(configfile)
        log.debug("Configuration file set to: %s", configfile)
        self.__resolved_config_file = config_file_path.resolve().as_posix()
        self._load_bare_config()

    def _load_bare_config(self):
        log.debug("Loading configuration file: %s", self.__resolved_config_file)
        config_file_path = PosixPath(self.__resolved_config_file)
        if config_file_path.exists():
            with config_file_path.open() as f:
                self.__bare_config_dict = json.load(f)
        else:
            self.__bare_config_dict = {}

    def save_bare_config_dict(self, bare_cfg):
        with PosixPath(self.__resolved_config_file).open('w') as f:
            f.write(json.dumps(bare_cfg, sort_keys=True, indent=4, separators=(',', ': ')))

    def get_bare_config_dict(self):
        return self.__bare_config_dict



# coding: utf-8

# Standard Libraries
import logging

# Dopplerr
from dopplerr.db import DopplerrDb
from dopplerr.descriptors.series import SeriesEpisodeUid
from dopplerr.notifications import emit_notifications
from dopplerr.notifications_types.series_subtitles_fetched import SubtitleFetchedNotification
from dopplerr.plugins.sonarr.filter import SonarrFilter
from dopplerr.tasks.download_subtitles import DownloadSubtitleTask
from dopplerr.tasks.manager import DopplerrTasksManager

log = logging.getLogger(__name__)


class DopplerrTask(object):

    def post_task(self, task):
        DopplerrTasksManager().post_task(task)


class TaskSonarrOnDownload(DopplerrTask):

    async def run(self, task) -> dict:
        content = task

        log.debug("Sonarr notification received: %r", content)
        res = await SonarrFilter().filter(content)
        if res.is_unhandled:
            # event has been filtered out
            return res.to_dict()

        if res.candidates:
            for candidate in res.candidates:
                DopplerrDb().insert_event("availability", "Available: {} - {}x{} - {} [{}].".format(
                    candidate.get("series_title"),
                    candidate.get("season_number"),
                    candidate.get("episode_number"),
                    candidate.get("episode_title"),
                    candidate.get("quality"),
                ))

        log.debug("Sonarr notification ok, posting background task")
        # processing ok, let's post our background task to the task queue
        self.post_task(self.task_sonarr_on_download_background(res))
        # asyncio.ensure_future(self.task_sonarr_on_download_background(res))
        res.successful("Request successfully posted")
        return res.to_dict()

    async def task_sonarr_on_download_background(self, res):
        log.debug("Starting task_sonarr_on_download_background")
        downloader = DownloadSubtitleTask()
        res = await downloader.run_and_wait(res)
        if not res.is_successful:
            log.debug("not successful, leaving background task")
            return res

        for episode_info in res.sonarr_episode_infos:
            await emit_notifications(SubtitleFetchedNotification(series_episode_info=episode_info))
            DopplerrDb().update_fetched_series_subtitles(
                series_episode_uid=SeriesEpisodeUid(
                    episode_info['tv_db_id'],
                    episode_info['season_number'],
                    episode_info['episode_number'],
                ),
                subtitles_languages=episode_info['subtitles_languages'],
                dirty=False)
        log.debug("Background task finished with result: %s", res.to_json())
        return res



# coding: utf-8

# Standard Libraries
import logging

# Third Party Libraries
from sanic import Blueprint
from sanic_transmute import add_route
from sanic_transmute import describe
from schematics.models import Model
from schematics.types import IntType
from schematics.types import ListType
from schematics.types import ModelType
from schematics.types import StringType

# Dopplerr
from dopplerr.db import DopplerrDb

log = logging.getLogger(__name__)


class Event(Model):
    timestamp = StringType()
    type = StringType()
    message = StringType()


class Events(Model):
    events = ListType(ModelType(Event))


class RecentSerie(Model):
    added_timestamp = StringType()
    series_title = StringType()
    season_number = IntType()
    episode_number = IntType()
    episode_title = StringType()
    quality = StringType()
    video_languages = IntType(required=False)
    subtitle_language = StringType()


class RecentSeries(Model):
    events = ListType(ModelType(RecentSerie))


bp = Blueprint('events', url_prefix="/api/v1/recent")


@describe(paths="/events/{num}", methods="GET")
async def recent_events_num(num: int = 10) -> Events:
    num = int(num)
    if num > 100:
        num = 100
    res = {"events": DopplerrDb().get_recent_events(num)}
    return res


add_route(bp, recent_events_num)


@describe(paths="/events/", methods="GET")
async def recent_events_10() -> Events:
    res = {"events": DopplerrDb().get_recent_events(10)}
    return res


add_route(bp, recent_events_10)


@describe(paths="/series/{num}", methods="GET")
async def recent_fetched_series_num(num: int = 10) -> RecentSeries:
    num = int(num)
    if num > 100:
        num = 100
    res = {"events": DopplerrDb().get_last_fetched_series(num)}
    return res


add_route(bp, recent_fetched_series_num)



# coding: utf-8

# Standard Libraries
import asyncio
import logging

# Dopplerr
from dopplerr.config import DopplerrConfig

log = logging.getLogger(__name__)


class PeriodicTask(object):
    job_id: str = None
    job_type = 'interval'
    job_default_kwargs = {'max_instances': 1}
    scheduler = None
    seconds: int = None
    minutes: int = None
    hours: int = None
    active = False
    _interrupted = False
    enable_cfg: str = None
    forced = False
    force_start_required = False

    def __init__(self):
        self.init()

    def init(self):
        pass

    async def run(self):
        try:
            if self.force_start_required:
                self.forced = True
            elif self.forced:
                log.debug("Forced execution already started, skip this periodic schedule")
                return
            self.active = True
            return await self._run()
        finally:
            self.active = False

    async def _run(self):
        raise NotImplementedError

    @property
    def _add_job_kwargs(self):
        kw = self.job_default_kwargs.copy()
        if self.seconds:
            kw['seconds'] = self.seconds
        if self.minutes:
            kw['minutes'] = self.minutes
        if self.hours:
            kw['hours'] = self.hours
        return kw

    @property
    def job(self):
        if self.scheduler:
            return self.scheduler.get_job(self.job_id)

    def add_job(self, scheduler):
        if self.enable_cfg is not None and not DopplerrConfig().get_cfg_value(self.enable_cfg):
            log.info("Do not enable job '%s', it is disabled by configuration '%s'", self.job_id,
                     self.enable_cfg)
            return
        self.scheduler = scheduler
        scheduler.add_job(
            self.run, self.job_type, id=self.job_id, replace_existing=True, **self._add_job_kwargs)

    @property
    def next_run_time(self):
        job = self.job
        if job:
            return self.job.next_run_time

    @property
    def next_run_time_iso(self):
        t = self.next_run_time
        if t:
            return t.isoformat()

    @property
    def interval(self):
        # yapf: disable
        return (
            (self.seconds if self.seconds else 0) +
            (self.minutes * 60 if self.minutes else 0) +
            (self.hours * 60 * 60 if self.hours else 0) +
            0)
        # yapf: enable

    @property
    def started(self):
        return self.scheduler

    def stop(self):
        self.scheduler = None

    @property
    def stopped(self):
        return not self.scheduler

    def interrupt(self):
        self._interrupted = True

    @property
    def interrupted(self):
        return self._interrupted

    async def force_start(self):
        log.debug("Force start job: %s", self.job_id)
        self.force_start_required = True
        asyncio.ensure_future(self.run())



# coding: utf-8

# Standard Libraries
import logging
import os
import sys
from pathlib import Path

# Dopplerr
from dopplerr import DOPPLERR_VERSION
from dopplerr.config import DopplerrConfig
from dopplerr.db import DopplerrDb
from dopplerr.logging import OutputType
from dopplerr.logging import setup_logging
from dopplerr.routes import listen
from dopplerr.status import DopplerrStatus
from dopplerr.tasks.subtasks.subliminal import SubliminalSubDownloader

log = logging.getLogger(__name__)


def main():
    outputtype = OutputType.PLAIN
    if "--debug-config" in sys.argv:
        default_level = logging.DEBUG
    else:
        default_level = logging.ERROR
    for i in range(len(sys.argv)):
        if sys.argv[i] == "--output-type":
            if i < len(sys.argv) and sys.argv[i + 1] == "dev":
                outputtype = OutputType.DEV
                break

    debug = default_level is logging.DEBUG
    setup_logging(debug=debug, outputtype=outputtype)
    log.debug("Initializing Dopplerr version %s...", DOPPLERR_VERSION)

    DopplerrConfig().find_configuration_values()
    log.debug("Current configuration: %s", DopplerrConfig().json(safe=True))

    debug = DopplerrConfig().get_cfg_value("general.verbose")
    output_type = DopplerrConfig().get_cfg_value("general.output_type")
    if output_type == 'dev':
        outputtype = OutputType.DEV
    elif output_type == 'plain':
        outputtype = OutputType.PLAIN
    else:
        raise NotImplementedError("Invalid output type: {!r}".format(output_type))

    log.debug("Applying configuration")
    custom_log_levels = [
        ("peewee", logging.DEBUG if debug else logging.ERROR),
        ("sanic", logging.INFO),
        ("cfgtree", logging.DEBUG if debug else logging.ERROR),
        ("apscheduler", logging.INFO),
        # Subliminal loggers
        ("chardet", logging.ERROR),
        ("dogpile", logging.ERROR),
        ("enzyme", logging.ERROR),
        ("rebulk.processors", logging.INFO),
        ("rebulk", logging.ERROR),
        ("subliminal.providers", logging.ERROR),
        ("subliminal.score", logging.ERROR),
        ("subliminal.subtitle", logging.ERROR),
        ("subliminal", logging.INFO),
        ("urllib3", logging.ERROR),
    ]
    setup_logging(
        outputtype=outputtype,
        debug=debug,
        logfile=DopplerrConfig().get_cfg_value("general.logfile"),
        custom_log_levels=custom_log_levels)
    log.info("Logging is set to %s", "verbose"
             if DopplerrConfig().get_cfg_value("general.verbose") else "not verbose")
    DopplerrStatus().refresh_from_cfg()
    # Backup configuration, now refresh_from_cfg has updated the version
    DopplerrConfig().save_configuration()

    log.info("Initializing Subtitle DopplerrDownloader Service")

    SubliminalSubDownloader.initialize_db()
    DopplerrStatus().sqlite_db_path = (
        Path(DopplerrConfig().get_cfg_value("general.configdir")) / "sqlite.db")
    reset_db = False
    if DopplerrStatus().has_minor_version_changed:
        log.warning("Major version change, dropping all databases")
        reset_db = True
    else:
        log.info("Previous version was %s, now: %s",
                 DopplerrStatus().previous_version,
                 DopplerrConfig().get_cfg_value("general.version"))
    log.debug("SQLite DB: %s", DopplerrStatus().sqlite_db_path.as_posix())
    DopplerrDb().init(DopplerrStatus().sqlite_db_path, reset_db=reset_db)
    DopplerrDb().create_tables()

    # change current work dir for subliminal work files
    os.chdir(DopplerrConfig().get_cfg_value("general.configdir"))

    DopplerrDb().insert_event("start", "dopplerr started")

    # main event loop (Asyncio behind)
    listen()

    logging.info("Clean stopping")
    DopplerrDb().insert_event("stop", "dopplerr stopped")

    return 0



# coding: utf-8

# Standard Libraries
import logging

# Third Party Libraries
from sanic import Blueprint
from sanic_transmute import APIException
from sanic_transmute import add_route
from sanic_transmute import describe

# Dopplerr
from dopplerr.plugins.sonarr.task import TaskSonarrOnDownload

log = logging.getLogger(__name__)

bp = Blueprint('notify', url_prefix="/api/v1")


@describe(paths="/notify/sonarr", methods=['POST'])
async def notify_sonarr(request):
    res = await TaskSonarrOnDownload().run(request.json)
    return res


add_route(bp, notify_sonarr)


@describe(paths="/notify", methods=['GET'])
async def notify_not_allowed():
    return APIException(
        "Method GET not allowed. "
        "Use POST with a JSON body with the right format", code=405)


add_route(bp, notify_not_allowed)



# coding: utf-8

# Standard Libraries
import logging

# Third Party Libraries
import aiohttp

# Dopplerr
from dopplerr.notifications._base import _NotificationBase

log = logging.getLogger(__name__)


class NotificationPushOver(_NotificationBase):

    __api_url = "https://api.pushover.net/1/messages.json"

    def __init__(self, token, user, registered_notifications):
        self.token = token
        self.user = user
        super(NotificationPushOver, self).__init__(registered_notifications)

    async def emit(self, notification_type, title, message):
        if not self.can_emit_notification_type(notification_type):
            log.debug("notification %s is ignored for pushover")
            return
        async with aiohttp.ClientSession() as session:
            d = {
                "token": self.token,
                "user": self.user,
                "message": message,
                "title": title,
            }
            async with session.post(self.__api_url, data=d) as result:
                response = await result.json()
                log.debug("PushOver response: %r", response)



# coding: utf-8

# Standard Libraries
import logging

# Third Party Libraries
from sanic import Blueprint
from sanic_transmute import add_route
from sanic_transmute import describe
from schematics.models import Model
from schematics.types import StringType

# Dopplerr
from dopplerr.config import DopplerrConfig

log = logging.getLogger(__name__)


class ConfigDir(Model):
    configdir = StringType()
    basedir = StringType()
    frontenddir = StringType()


bp = Blueprint('config', url_prefix="/api/v1")


@describe(paths="/config/general/dirs")
async def config_directories() -> ConfigDir:
    return {
        "configdir": DopplerrConfig().get_cfg_value("general.configdir"),
        "basedir": DopplerrConfig().get_cfg_value("general.basedir"),
        "frontenddir": DopplerrConfig().get_cfg_value("general.frontenddir"),
    }


add_route(bp, config_directories)



# coding: utf-8

# Standard Libraries
import logging

# Dopplerr
from dopplerr.singleton import singleton

log = logging.getLogger(__name__)


@singleton
class DopplerrDownloader(object):

    def process_fullscan(self, _request):
        log.debug("Processing full scan of missing subtitle files...")
        res = {
            'status': 'unprocessed',
            'message': 'not implemented yet!',
        }
        # TODO: inspiration
        #   https://gist.github.com/alexsavio/9299716
        return res



# coding: utf-8

# Standard Libraries
import logging
import os
from pathlib import PosixPath

# Third Party Libraries
import pkg_resources
from cfgtree.cfgtree import EnvironmentConfig
from cfgtree.storages import JsonFileConfigStorage
from cfgtree.types import BoolCfg
from cfgtree.types import ConfigFileCfg
from cfgtree.types import DirNameCfg
from cfgtree.types import HardcodedCfg
from cfgtree.types import IntCfg
from cfgtree.types import ListOfStringCfg
from cfgtree.types import MultiChoiceCfg
from cfgtree.types import PasswordCfg
from cfgtree.types import SingleChoiceCfg
from cfgtree.types import StringCfg
from cfgtree.types import UserCfg

# Dopplerr
from dopplerr.singleton import singleton

log = logging.getLogger(__name__)
DEFAULT_PORT = 8086


def _find_frontend_data():
    installed_data_frontend = pkg_resources.resource_filename(__name__, 'frontend')
    if PosixPath(installed_data_frontend).exists():
        log.debug("Found local frontend path: %s", installed_data_frontend)
        return installed_data_frontend
    setup_py = pkg_resources.resource_filename(__name__, "main.py")
    dev_env_frontend_dist = PosixPath(setup_py).parent.parent / "frontend" / "dist"
    if dev_env_frontend_dist.exists():
        log.debug("Found dev local frontend path: %s", dev_env_frontend_dist)
        return str(dev_env_frontend_dist)
    return None


class DopplerrJsonConfigFile(JsonFileConfigStorage):
    json_configstorage_environ_var_name = "DOPPLERR_COMMON_CONFIG_FILE"
    json_configstorage_long_param_name = "--configfile"
    json_configstorage_short_param_name = "-g"
    json_configstorage_default_filename = "config.json"


@singleton
class DopplerrConfig(EnvironmentConfig):

    environ_var_prefix = "DOPPLERR_"
    config_storage = DopplerrJsonConfigFile()

    cfgtree = {
        "configfile": ConfigFileCfg(l="--configfile", h="Config directory"),
        "debug_config": BoolCfg(l="--debug-config", h="Show logs before configuration load"),
        "general": {
            "basedir":
                DirNameCfg(s="-b", d=os.getcwd(), h='Base directory'),
            "configdir":
                DirNameCfg(s="-c", d=os.getcwd(), h="Config directory"),
            "frontenddir":
                DirNameCfg(s="-f", d=_find_frontend_data(), r=True, h="Frontend directory"),
            "verbose":
                BoolCfg(s='-v', l="--verbose", h='Enable verbose output logs'),
            "output_type":
                SingleChoiceCfg(
                    l="--output-type",
                    h="Output log type",
                    choices=["quiet", "plain", "dev"],
                    d="plain"),
            "logfile":
                StringCfg(s="-l", h='Output log to file'),
            "mapping":
                ListOfStringCfg(
                    s='-m',
                    h=("Map root folder of tv/anime/movie to another name.\n"
                       "Ex: series are mounted on a docker image as /tv but \n"
                       "on the other system it is under /video/Series. In this \n"
                       "case use '--basedir /video --mapping tv=Series,movies=Movies'\n"
                       "Please enter trivial mapping as well:\n"
                       "   '--mapping tv=tv,movies=movies'")),
            "port":
                IntCfg(s='-p', d=DEFAULT_PORT, h='The port to listen on'),
            "no_color":
                BoolCfg(h="Disable color in logs"),
            "version":
                HardcodedCfg(),
        },
        "subliminal": {
            "languages": ListOfStringCfg(),
            "addic7ed": {
                "enabled": BoolCfg(h="Enable addic7ed"),
                "user": UserCfg(h="addic7ed username"),
                "password": PasswordCfg(h="addic7ed password"),
            },
            "legendastv": {
                "enabled": BoolCfg(h="Enable legendastv"),
                "user": UserCfg(h="legendastv username"),
                "password": PasswordCfg(h="legendastv password"),
            },
            "opensubtitles": {
                "enabled": BoolCfg(h="Enable opensubtitles"),
                "user": UserCfg(h="opensubtitles username"),
                "password": PasswordCfg(h="opensubtitles password"),
            },
            "subscenter": {
                "enabled": BoolCfg(h="Enable subscenter"),
                "user": UserCfg(h="subscenter username"),
                "password": PasswordCfg(h="subscenter password"),
            },
        },
        "notifications": {
            "pushover": {
                "enabled":
                    BoolCfg(h="Enable pushover"),
                "user":
                    UserCfg(h="pushover username"),
                "token":
                    PasswordCfg(h="pushover password"),
                "registered_notifications":
                    MultiChoiceCfg(h="Notifications", choices=["fetched"], d=["fetched"]),
            }
        },
        "scanner": {
            "enable": BoolCfg(h="Enable periodic disc scanner", d=False),
            "interval_hours": IntCfg(h="Refresh interval (in hours)", d=6),
        }
    }



# coding: utf-8

# Standard Libraries
import asyncio
import logging

# Third Party Libraries
from apscheduler.schedulers.asyncio import AsyncIOScheduler

# Dopplerr
from dopplerr.singleton import singleton
from dopplerr.tasks.disk_scanner import DiskScanner
from dopplerr.tasks.download_subtitles import DownloadSubtitleTask

log = logging.getLogger(__name__)


@singleton
class DopplerrTasksManager(object):
    background_tasks = 0
    apscheduler = None

    def post_task(self, task):
        """
        TODO: transfor this simple wrapper arround `ensure_future ` into a real queue management.
        """

        async def wrap_task(task):
            try:
                self.background_tasks += 1
                return await task
            finally:
                self.background_tasks -= 1

        asyncio.ensure_future(wrap_task(task))

    def start(self):
        DownloadSubtitleTask().start()
        self.apscheduler = AsyncIOScheduler()
        DiskScanner().add_job(self.apscheduler)
        self.apscheduler.start()

    def stop(self):
        DownloadSubtitleTask().stop()
        self.apscheduler.shutdown(False)

    def status(self):
        return {
            'background_tasks': self.background_tasks,
            'subtitle_downloader': {
                'started': 1 if DownloadSubtitleTask().started else 0,
                'active': 1 if DownloadSubtitleTask().active else 0,
            },
            'disc_scanner': {
                'started': 1 if DiskScanner().started else 0,
                'active': 1 if DiskScanner().active else 0,
                'interval_hours': DiskScanner().interval_hours,
                'next_run_time': DiskScanner().next_run_time_iso,
            }
        }



# coding: utf-8
"""
Singleton class definition.

Usage:

    @singleton
    class MySingletonClass(object):
        def __init__(self):
            pass
        ...
        def a_function(self):
            pass

Later in the code:

    from module.of.MySingleton import MySingletonClass
    def anyFunction(...):
        ...
        MySingletonClass().a_function(...)


"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

# Standard Libraries
import types

# pylint: disable=invalid-name


class __Singleton(object):

    """
    A non-thread-safe helper class to ease implementing singletons.

    This should be used as a decorator -- not a metaclass -- to the
    class that should be a singleton.

    The decorated class can define one `__init__` function that
    takes only the `self` argument. Other than that, there are
    no restrictions that apply to the decorated class.

    To get the singleton instance, use the `instance` method. Trying
    to use `__call__` will result in a `TypeError` being raised.

    Limitations: The decorated class cannot be inherited from.
    """

    def __init__(self, decorated):
        self._decorated = decorated

    def instance(self, *args, **kwargs):
        """
        Return the singleton instance.

        Upon its first call, it creates a new instance of the decorated class and calls its
        `__init__` method.

        On all subsequent calls, the already created instance is returned.
        """
        # Do not use a test here for performance sake. The first call has the exception penalty
        try:
            return self._instance
        except AttributeError:
            # pylint: disable=attribute-defined-outside-init
            self._instance = self._decorated(*args, **kwargs)

            # pylint: enable=attribute-defined-outside-init

            def unload(inst):
                # pylint: disable=protected-access
                inst.__singleton.unload()
                # pylint: enable=protected-access

            # Magically bind "unload" as a method
            self._instance.unload = types.MethodType(unload, self._instance)
            # pylint: disable=protected-access
            self._instance.__singleton = self
            # pylint: enable=protected-access
            return self._instance

    def __call__(self, *args, **kwargs):
        return self.instance(*args, **kwargs)

    def unload(self):
        if hasattr(self, "_instance"):
            delattr(self, "_instance")


singleton = __Singleton
# pylint: enable=invalid-name



# coding: utf-8

# Standard Libraries
import logging
import re

# Third Party Libraries
from cfgtree import LOGGER_NAME

log = logging.getLogger(LOGGER_NAME)

# tells flake8 to ignore complexity check for this file
# flake8: noqa


def get_node_by_xpath(mapping, xpath, default=None, ignore_errors=False,
                      handle_list_selector=False):
    '''Return the node pointed to by xpath from mapping.

    Args:
        mapping: nested dictionary.
        xpath: string-like selector.
        default: default value if the attribute doesn't exist.
        ignore_errors: if True, pass silently if the xpath is invalid.
        handle_list_selector: allow to support list element selector

    Example:

    >>> tree = {'level1': {'level2': {'level3': 'bottom'}}}
    >>> get_node_by_xpath(tree, 'level1.level2.level3') == 'bottom'
    True

    '''
    if not isinstance(mapping, dict):
        if not ignore_errors:
            raise KeyError("Mapping is not dictionary: {!r}".format(mapping))
        return default
    for segment in xpath.split('.'):
        if not mapping:
            if not ignore_errors:
                raise KeyError("Empty mapping, but need to access to '{}'".format(xpath))
            return default
        if segment not in mapping:
            if handle_list_selector and '[' in segment:
                re_subselector = re.compile(r"(.*)\[(\d+)\]$")
                m = re_subselector.match(segment)
                if m:
                    key = m.group(1)
                    index = int(m.group(2))
                    if key not in mapping:
                        if not ignore_errors:
                            raise KeyError("Invalid '{}' index selector: '{}' does not match "
                                           "anything. Available keys: {!r}".format(
                                               xpath, key, list(mapping.keys())))
                        return default
                    items = mapping[key]
                    if not isinstance(items, list):
                        if not ignore_errors:
                            raise KeyError("Invalid '{}' selector: '{}' is not a list, is: {}"
                                           .format(xpath, key, type(items)))
                        return default
                    if len(items) <= index:
                        if not ignore_errors:
                            raise KeyError("Invalid '{}' selector: item index '{}' of '{}' is "
                                           "outside of the list boundaries. Length is: {}".format(
                                               xpath, index, key, len(items)))
                        return default
                    mapping = items[index]
                    continue
            elif not ignore_errors:
                raise KeyError("Invalid '{}' selector: '{}' doesn't match "
                               "anything. Available keys: {!r}".format(
                                   xpath, segment, list(mapping.keys())))
            return default
        mapping = mapping[segment]
    return mapping


def set_node_by_xpath(mapping, xpath, value, extend=False, setter_attr=None):
    '''Set the node pointed to by xpath from mapping.

    Args:
        mapping: nested dictionary.
        xpath: string-like selector.
        value: value to set.
        extend: if True, create the nested structure if it doesn't exist,
            otherwise, raise an exception.
        setter_attr: use a special setter method attribute in mapping, instead of replacing
                     the node by the new value (note: do not use a property setter attribute)

    Example:

    >>> tree = {'level1': {'level2': {'level3': 'bottom'}}}
    >>> set_node_by_xpath(tree, 'level1.level2.level3', 'bottom')

    '''
    segments = xpath.split('.')
    attrname = segments.pop()
    for segment in segments:
        if segment not in mapping:
            if not extend:
                raise KeyError("Invalid '{}' selector: '{}' doesn't match "
                               "anything.".format(xpath, segment))
            mapping[segment] = {}
        mapping = mapping[segment]
    if setter_attr:
        # setter attribute defined, calling this setter
        setter = getattr(mapping[attrname], setter_attr)
        setter(value)
    else:
        mapping[attrname] = value


def delete_node_by_xpath(mapping, xpath, ignore_errors=False):
    '''Delete the node pointed to by xpath from mapping.

    Args:
        mapping: nested dictionary.
        xpath: string-like selector.
        ignore_errors: if True, pass silently if the node doesn't exist,
            otherwise, raise an exception.

    Example:

    >>> tree = {'level1': {'level2': {'level3': 'bottom'}}}
    >>> delete_node_by_xpath(tree, 'level1.level2')
    >>> tree
    {'level1': {}}

    '''
    segments = xpath.split('.')
    attrname = segments.pop()
    for segment in segments:
        if segment not in mapping:
            if ignore_errors:
                return
            raise KeyError("Invalid '{}' selector: '{}' doesn't match "
                           "anything.".format(xpath, segment))
        mapping = mapping[segment]
    return mapping.pop(attrname, None)



# coding: utf-8

# Standard Libraries
import asyncio
import concurrent
import functools
import logging

log = logging.getLogger(__name__)


class ThreadedTask(object):
    worker_threads_num = 1

    def __init__(self):
        self.executors = concurrent.futures.ThreadPoolExecutor(max_workers=self.worker_threads_num)

    async def _run_in_thread(self, func, *args, **kwargs):
        event_loop = asyncio.get_event_loop()
        res = await event_loop.run_in_executor(self.executors,
                                               functools.partial(func, *args, **kwargs))
        return res



# coding: utf-8

# Standard Libraries
import logging

# Dopplerr
from dopplerr.config import DopplerrConfig
from dopplerr.plugins.sonarr.response import SonarrOnDownloadResponse
from dopplerr.request_filter import _FilterBase
from dopplerr.response import UnhandledResponse

log = logging.getLogger(__name__)


class SonarrFilter(_FilterBase):

    async def filter(self, request):
        # probably request_event
        low_request = self.lowerize_dick_keys(request)
        eventtype = low_request.get("eventtype")
        if eventtype == "Download":
            res = await self.process_on_download(request)
            return res
        return UnhandledResponse("sonarr", "Ignoring Sonarr event type: {!r}".format(eventtype))

    async def process_on_download(self, request):
        log.debug("Processing Sonarr's 'on downloaded' event")
        res = SonarrOnDownloadResponse()
        res.processing()
        low_request = self.lowerize_dick_keys(request)
        low_series = self.lowerize_dick_keys(low_request.get("series", {}))
        root_dir = low_series.get("path")
        series_title = low_series.get("title")
        tv_db_id = low_series.get("tvdbid")
        if not root_dir:
            res.failed("Empty Series Path")
            return res
        root_dir = self.appy_path_mapping(root_dir)
        log.debug("Root folder: %s", root_dir)
        log.debug("Reconstructing full media path with basedir '%s'",
                  DopplerrConfig().get_cfg_value("general.basedir"))

        def concat_path(str_a, str_b):
            if not str_a.endswith('/'):
                str_a += '/'
            if str_b.startswith('/'):
                str_b = str_b[1:]
            str_a += str_b
            return str_a

        root_dir = concat_path(DopplerrConfig().get_cfg_value("general.basedir"), root_dir)
        basename = root_dir
        log.info("Searching episodes for serie '%s' in '%s'", series_title, root_dir)
        res.processing("listing candidates")
        for episode in low_request.get("episodes", []):
            low_episode = self.lowerize_dick_keys(episode)
            basename = low_episode.get("scenename", "")
            episode_title = low_episode.get("title", "")
            season_number = low_episode.get("seasonnumber", "")
            episode_number = low_episode.get("episodenumber", "")
            quality = low_episode.get("quality", "")
            log.debug("Candidate: episode '%s' with base filename '%s'", episode_title, basename)
            res.candidates.append({
                "series_title": series_title,
                "tv_db_id": tv_db_id,
                "episode_title": episode_title,
                "root_dir": root_dir,
                "scenename": basename,
                "season_number": season_number,
                "episode_number": episode_number,
                "quality": quality,
            })
        res.processing("candidates listed")
        return res



# coding: utf-8

# Standard Libraries
import logging

log = logging.getLogger(__name__)


class _NotificationTypeBase(object):
    pass


class _SeriesNotificationBase(_NotificationTypeBase):
    notification_type: str = None
    notification_title: str = None
    series_episode_info = None

    def __init__(self, series_episode_info):
        self.series_episode_info = series_episode_info



# coding: utf-8

# Standard Libraries
import logging
import os
import sys
from enum import Enum
from functools import partial
from io import StringIO
from logging.handlers import RotatingFileHandler
from subprocess import PIPE
from subprocess import check_output

log = logging.getLogger(__name__)

# flake8: noqa
# pylint: disable=too-many-locals,bare-except


class OutputType(Enum):
    PLAIN = 1
    DEV = 2
    JSON = 3


g_file_handler = None
g_stream_handler = None


def setup_logging(outputtype=OutputType.PLAIN,
                  debug=False,
                  unbuffered=False,
                  logfile=None,
                  dev_allow_colors=True,
                  dev_split=True,
                  dev_module_verbose=True,
                  dev_force_fmt=None,
                  dev_force_no_tty=None,
                  custom_log_levels=None):
    '''
    :param outputtype:        one of the following:
                                - OutputType.PLAIN: plain log output (all in stdout)
                                - OutputType.DEV: developer-friendly, multi-columns, colored output
                                                  (if `colorlog` package is available and
                                                  `dev_allow_colors` set to True)
                                - OutputType.JSON: output in json format, 1 json-chunk per line
    :param debug:             configure log level. Can take the following values:
                                - False: logs will be in level logging.INFO
                                - True: logs will be in level logging.DEBUG
                                - logging.[DEBUG|INFO|WARNING|ERROR|FATAL]: set log level manually
    :param unbuffered:        ask the stdout to avoid buffering to send logs as fast as possible to
                              the caller script.
                              Only available when not inside a TTY, especially for pipeline of logs.
                              This also trigger the split of outputs into stdout and stderr:
                                  >=logging.ERROR, logs go to stderr
                                  < logging.ERROR, logs go to stdout
    :param logfile:           enable output of logs to a file
    :param custom_log_levels  list of tuple (logger_name, loglevel) to change loglevel on diferent
                              loggers.

                              Example:
                                # Hide debug and info levels from the `urllib3` module
                                setup_logging(
                                  debug=logging.DEBUG,
                                  custom_log_levels = [
                                      ("urllib3", logging.ERROR),
                                  ]
                                  ...
                                )

    The rest of the parameters allow to fine-tune developer-mode:

    :param dev_allow_colors:   allow log coloration if `colorlog` is installed
    :param dev_split:          preserve log headers for multiline strings, and adapt to the current
                               terminals witdh if applicable
    :param dev_module_verbose: add module (file, line number) to each log line
    :param dev_force_fmt:      overwrite the formatter
    :param dev_force_no_tty:   force the usage of the no TTY formatter
    '''
    ColoredFormatter = False  # pylint: disable=invalid-name
    if dev_allow_colors:
        try:
            # Note: colorlog is not a 'requirement' of your application (and shouldn't!),
            # so it might not be installed.
            # In this case, normal logging formatter is used.
            # But if available, let's print plentful of colors!
            #
            # To install it, simply do:
            #
            #    pip install colorlog
            #
            from colorlog import ColoredFormatter
        except:
            pass

    # this logging config will output debug and info to stdout and warning, error, and critical to
    # stderr, so that consumer can color them if needed
    # logs with `level > stderr_threshold` will go to `stderr`,
    # logs with `level < stderr_threshold` will go to `stdout`
    stderr_threshold = logging.WARNING
    # logs with `level < stdout_level` will be hidden
    stdout_level = logging.DEBUG

    class InfoFilter(logging.Filter):

        def filter(self, record):
            return record.levelno < stderr_threshold

    class SplitFormatterMixin(object):

        '''
        Magic Formatter that gracefully handle multiline logs, ie it split the multiline string
        and add each log independently.

        This handles two spliting:

        - if the terminal width is found, try to fit the output in the terminal by splitting the
          string arbitrarily
        - if the record message is multiline, split it and print prefix (loglevel, date,...)
          independently

        Note this only impacts the display to stdout handler, not the record storage, so
        if these logs are sent to another handler (for instance to splunk handler), the log
        will not be split
        '''

        def format(self, record):
            record.message = record.getMessage()
            record.msg = record.message
            if record.args:
                record.args = ()
            if self.usesTime():
                record.asctime = self.formatTime(record, self.datefmt)
            multiline_message = record.message
            formatted_lines = []
            if not self.split:
                s = self._get_formated_line(record)
                formatted_lines.append(s)
            else:
                for line in multiline_message.split("\n"):
                    if self.term_width < 1:
                        sublines = [line]
                    else:
                        sublines = [
                            l for l in iter(partial(StringIO(line).read, int(self.term_width)), '')
                        ]
                    if not sublines:
                        # If the string is empty, still add an empty record in the output
                        record.message = ""
                        record.msg = ""
                        s = self._get_formated_line(record)
                        formatted_lines.append(s)
                    else:
                        for subline in sublines:
                            record.message = subline
                            record.msg = subline
                            s = self._get_formated_line(record)
                            formatted_lines.append(s)
            return "\n".join(formatted_lines)

        def _get_formated_line(self, record):
            s = self._get_record_with_fmt(record)
            if record.exc_info:
                # Cache the traceback text to avoid converting it multiple times
                # (it's constant anyway)
                if not record.exc_text:
                    record.exc_text = self.formatException(record.exc_info)
            if record.exc_text is not None:
                if s and s[-1:] != "\n":
                    s = s + "\n"
                try:
                    s = s + record.exc_text
                except UnicodeError:
                    # Sometimes filenames have non-ASCII chars, which can lead
                    # to errors when s is Unicode and record.exc_text is str
                    # See issue 8924.
                    # We also use replace for when there are multiple
                    # encodings, e.g. UTF-8 for the filesystem and latin-1
                    # for a script. See issue 13232.
                    s = s + record.exc_text.decode(sys.getfilesystemencoding(), 'replace')
            return s

    class SplitFormatter(SplitFormatterMixin, logging.Formatter):

        def __init__(self, fmt=None, datefmt=None, term_width=-1, split=True):
            super(SplitFormatter, self).__init__(fmt, datefmt)
            self.term_width = term_width
            self.split = split

        def _get_record_with_fmt(self, record):
            return self._fmt % record.__dict__

    if ColoredFormatter:

        class SplitColorFormatter(SplitFormatterMixin, ColoredFormatter):

            def __init__(self, fmt=None, datefmt=None, term_width=-1, split=True, log_colors=None):
                super(SplitColorFormatter, self).__init__(fmt, datefmt, log_colors=log_colors)
                self.term_width = term_width
                self.split = split

            def _get_record_with_fmt(self, record):
                return ColoredFormatter.format(self, record)

    align_level_width = 8
    extra_char_width = 3
    if outputtype == OutputType.PLAIN:
        fmt_color_str = fmt_nocolor_str = "%(message)s"
    elif outputtype == OutputType.DEV:
        if dev_force_fmt:
            fmt_nocolor_str = dev_force_fmt
        else:
            if dev_module_verbose:
                fmt_module_nocolor_str = "[%(name)-37.37s] "
                fmt_module_color_str = ("[%(yellow)s%(name)-37.37s%(reset)s] ")
            else:
                fmt_module_nocolor_str = ""
                fmt_module_color_str = ""
            fmt_debug_nocolor_str = (
                '%(asctime)19.19ss ' + fmt_module_nocolor_str + '%(levelname)7s: %(message)s')
            fmt_debug_color_str = (
                '%(blue)s%(asctime)19.19s%(reset)s ' + fmt_module_color_str +
                '%(log_color)s%(levelname)-7s%(reset)s | ' + '%(log_color)s%(message)s%(reset)s')

            fmt_simple_nocolor_str = '%(levelname)-7s - ' + fmt_module_nocolor_str + '%(message)s'
            fmt_simple_color_str = (
                '%(log_color)s' + fmt_module_color_str + '[%(levelname).1s]%(reset)s %(log_color)s'
                '%(message)s%(reset)s')

            if debug is True:
                fmt_color_str = fmt_debug_color_str
                fmt_nocolor_str = fmt_debug_nocolor_str
            else:
                fmt_color_str = fmt_simple_color_str
                fmt_nocolor_str = fmt_simple_nocolor_str
    elif outputtype == OutputType.JSON:
        raise NotImplementedError()

    if debug is True:
        default_level = logging.DEBUG
    elif debug is False:
        default_level = logging.INFO
    else:
        assert debug in [
            logging.CRITICAL,
            logging.ERROR,
            logging.WARNING,
            logging.INFO,
            logging.DEBUG,
        ], "invalid debug (neither True/False neither a logging level)"
        default_level = debug

    logging.basicConfig(stream=sys.stdout, level=default_level, format=fmt_nocolor_str)
    if custom_log_levels:
        for logger_name, loglevel in custom_log_levels:
            logger = logging.getLogger(logger_name)
            logger.setLevel(loglevel)

    date_fmt_string = None
    # Replace the default formatter because it is buggy
    root = logging.getLogger()
    # Do *not* replace the formatter in quiet mode since we want to bare output

    # force log level on root, in case of multiple call of basicConfig, if the first level
    # was higher than the second, the root log level is not not updated
    root.setLevel(default_level)
    assert log.getEffectiveLevel() == default_level, "invalid log level set on root !"

    term_width = -1
    if dev_split:
        # Try to retrieve the terminal width:
        term_width = -1
        try:
            # redirecting the stderr to an unused PIPE, to avoid the following issue:
            #     'stty: standard input: Invalid argument'
            # when executed from a step ShellCommand
            data = check_output(['stty', 'size'], stderr=PIPE)
            _, columns = data.split()
            if columns > 0:
                term_width = columns
        except:
            pass
        term_width = (int(term_width) - align_level_width - extra_char_width)
    # disable terminal spliting for the moment
    term_width = -1

    if dev_force_no_tty:
        is_tty = False
    else:
        is_tty = sys.stdout.isatty()

    if not is_tty:
        # We are in Buildbot or in a pipe, don't use colorlog!
        ColoredFormatter = None  # pylint: disable=invalid-name
    if not ColoredFormatter:
        formatter = SplitFormatter(fmt_nocolor_str, term_width=term_width, split=True)
    else:
        formatter = SplitColorFormatter(
            fmt_color_str,
            date_fmt_string,
            term_width=term_width,
            split=True,
            log_colors={
                'DEBUG': 'cyan',
                'INFO': 'green',
                'WARNING': 'yellow',
                'ERROR': 'red',
                'CRITICAL': 'red,bg_white',
            })

    global g_file_handler
    global g_stream_handler

    if unbuffered and not is_tty:
        root.handlers = []
        # Make stdout unbuffered, so that we get output asap to buildbot log
        sys.stdout = os.fdopen(sys.stdout.fileno(), 'w', 0)
        sys.stderr = os.fdopen(sys.stderr.fileno(), 'w', 0)
        h1 = logging.StreamHandler(sys.stdout)
        h1.setLevel(stdout_level)
        h1.setFormatter(formatter)
        h1.addFilter(InfoFilter())
        h2 = logging.StreamHandler(sys.stderr)
        h2.setFormatter(formatter)
        h2.setLevel(stderr_threshold)

        # Don't print useless info about "New connection" + security warnings
        logging.getLogger("requests.packages.urllib3.connectionpool").setLevel(logging.ERROR)

        logging.getLogger().addHandler(h1)
        logging.getLogger().addHandler(h2)
        logging.getLogger().setLevel(logging.NOTSET)
    else:
        root.handlers[0].setFormatter(formatter)

    if logfile:
        log.debug("Also output logs to file: %s", logfile)
        file_fmt = " :: ".join(["%(asctime)s", "%(levelname)s", "%(name)s", "%(message)s"])
        file_formatter = logging.Formatter(file_fmt)
        file_handler = RotatingFileHandler(str(logfile), 'a', 5 * 1024 * 1024, 1)
        file_handler.setFormatter(file_formatter)
        if g_file_handler:
            root.removeHandler(g_file_handler)
        root.addHandler(file_handler)
        g_file_handler = file_handler
